import numpy as np
import sys, getopt, os, commands
import sklearn.datasets as sk_data

# read data with missing values
# assumes all known entries are non-negative
def readIncompleteMtx(filename,noedge):
    try:
        sys.stderr.write('reading input from file: %s\n'%filename)
        x = np.genfromtxt(filename,dtype='float',delimiter=';',filling_values=noedge)
        M=np.matrix(x).astype('float')
        m = M.min()
        if m<noedge:
            sys.stderr.write('WARNING: matrix is not non-negative!\n')
            sys.stderr.write( 'WARNING: matrix is not a Gram-matrix\n')
            sys.stderr.write('WARNING: min non-missing value: %f\n\n'%m)
        if np.array_equal(np.transpose(M),M) == False:
            sys.stderr.write('WARNING: matrix is not symmetric!\n\n')
        return M
    except:
        sys.stderr.write('cannot find file or read data filename: %s\n'%filename)
        M = np.matrix('-1')
        return M

def writeMtxToTxt(M,output_file):
    np.savetxt(output_file,M,delimiter=';',newline='\n')

def partialDiagMatrix(M,k):
    Mdiag = np.diag(np.diag(M))
    for i in range(1,k+1):
        Mdiag = Mdiag + np.diag(np.diag(M,i),i)+  np.diag(np.diag(M,-i),-i)
    return Mdiag

#generates size nxn (int) mtx with effective rk k (int) and noise level noise (float)
def genLowRankGaussianMtx(n,k, noise):
    M = sk_data.make_low_rank_matrix(n,n,k,noise)
    #minM = np.min(M)
    #if minM<0:
    #    M = M-(minM*np.ones(np.shape(M)))
    Mt = np.transpose(M)
    Ms = np.dot(M,Mt)
    return Ms
