import numpy as np
import networkx as nx
import PriorityDictionary as pd
import sys

from sklearn.utils import check_random_state

# mindegree algorithm for elimination game
# noedge is the fill value for missing values in M
# Returns minimum elimination ordering, which is perfict elim ord iff M is chordal
def minDegreeAlgorithm(M, seed=None):
    G = make_mask_graph(M)
    # keep nodes in increasing order of remaining degree
    degree_pd = pd.priorityDictionary()
    # prng.shuffle(nodes)
    for i in G:
        # count number of known values in row i
        degree_pd[i] = len(G[i])
    # list with perfect elimination order of nodes
    MEO = []
    # add items to MEO always starting from smallest remaining degree
    for mindeg_node in degree_pd:
        # add node with smallest number of downstream neighbors to MEO
        MEO.append(mindeg_node)
        # if np.shape(np.nonzero((M[mindeg,:]>noedge))) == (1,1):
        #     continue
        # if np.count_nonzero((M[mindeg,:]!=noedge)) <=1: continue
        # get neighborlist of node with min degree
        # a = np.nonzero((M[mindeg,:]!=noedge))[0]
        neighbors = set(G[mindeg_node])
        G.remove_node(mindeg_node)
        for neighbor in neighbors:
            new_neighbors = neighbors - set([neighbor]) - set(G[neighbor])
            for new_neighbor in new_neighbors:
                G.add_edge(neighbor, new_neighbor)
            degree_pd[neighbor] = len(G[neighbor])
    return MEO


def perfect_order(M, seed=None):
    G = make_mask_graph(M)
    # keep nodes in increasing order of remaining degree
    degree_pd = pd.priorityDictionary()
    # prng.shuffle(nodes)
    for i in G:
        # count number of known values in row i
        degree_pd[i] = len(G[i])
    # list with perfect elimination order of nodes
    MEO = []
    # add items to MEO always starting from smallest remaining degree
    for mindeg_node in degree_pd:
        # add node with smallest number of downstream neighbors to MEO
        MEO.append(mindeg_node)
        # if np.shape(np.nonzero((M[mindeg,:]>noedge))) == (1,1):
        #     continue
        # if np.count_nonzero((M[mindeg,:]!=noedge)) <=1: continue
        # get neighborlist of node with min degree
        # a = np.nonzero((M[mindeg,:]!=noedge))[0]
        neighbors = set(G[mindeg_node])
        G.remove_node(mindeg_node)
        for neighbor in neighbors:
            new_neighbors = neighbors - set([neighbor]) - set(G[neighbor])
            for new_neighbor in new_neighbors:
                G.add_edge(neighbor, new_neighbor)
                G.add_edge(new_neighbor, neighbor)
            degree_pd[neighbor] = len(G[neighbor])
    return MEO


def min_deficiency_order(M, sample=False, random_state=None):
    prng = check_random_state(random_state)
    G = make_mask_graph(M)
    n = len(G)
    deficiencies = pd.priorityDictionary()
    for i in G:
        deficiencies[i] = deficiency(i, G)
    PEO = []

    while len(PEO) < n:
        min_def = float('inf')
        min_def_nodes = []
        for node in deficiencies:
            v = deficiencies[node]
            if v <= min_def:
                min_def = v
                min_def_nodes.append(node)
            else:
                break
        # if the last node has deficiency larger than min, put it back
        if v > min_def:
            deficiencies[node] = v
        if sample:
            picked_index = prng.randint(len(min_def_nodes))
        else:
            picked_index = 0
        picked_node = min_def_nodes[picked_index]
        for i in range(len(min_def_nodes)):
            if i != picked_index:
                deficiencies[min_def_nodes[i]] = min_def

        PEO.append(picked_node)
        neighbors = list(set(G[picked_node]) - {picked_node})
        G.remove_node(picked_node)
        for i, neighbor_i in enumerate(neighbors):
            for j in range(i + 1, len(neighbors)):
                neighbor_j = neighbors[j]
                G.add_edge(neighbor_i, neighbor_j)
        for neighbor in neighbors:
            deficiencies[neighbor] = deficiency(neighbor, G)
    return PEO


def make_mask_graph(mtx):
    G = nx.Graph()
    n = mtx.shape[0]
    for node in range(n):
        G.add_node(node)
    edges = [(i, j)
             for i in range(n)
             for j in range(n)
             if i > j and not np.isnan(mtx[i, j])]
    G.add_edges_from(edges)
    # for i in range(n):
        # for j in range(n):
            # if i >= j:
                # continue
            # if not np.isnan(mtx[i, j]):
                # G.add_edge(i, j)
    return G


def original_mindegree(M):
    sys.stderr.write('Computing MEO start \n')
    n = len(M)
    # keep nodes in increasing order of remaining degree
    degrees = pd.priorityDictionary()
    for i in range(n):
        # count number of known values in row i
        degrees[i] = np.count_nonzero(~np.isnan(M[i, :]))
        print i, degrees[i]
    # list with perfect elimination order of nodes
    MEO = []

    # add items to MEO always starting from smallest remaining degree
    for mindeg in degrees:
        #add node with smallest number of downstream neighbors to MEO
        MEO.append(mindeg)
        #if np.shape(np.nonzero((M[mindeg,:]>noedge))) == (1,1):
        #     continue
        if np.count_nonzero(~np.isnan(M[mindeg,:])) <=1: continue
        #get neighborlist of node with min degree
        a = np.nonzero(~np.isnan(M[mindeg, :]))[0]
        print a
        for neigh in a:
            #skip neighbors that are already handled
            if neigh == mindeg: continue
            if neigh in MEO:continue
            #reduce downstream degree  of neighbors by 1
            degrees[neigh] -=1
    sys.stderr.write('Computing MEO end \n')
    return MEO


def deficiency(node, G):
    deficiency = 0
    neighbors = list(set(G[node]))
    for i, node_i in enumerate(neighbors):
        for j in range(i + 1, len(neighbors)):
            if neighbors[j] not in G[node_i]:
                deficiency += 1
    return deficiency
