import numpy as np
import sys


# return square principal submatrix, with top right corner indexed by i,j
# returns only a view of the submatrix
def getSubMtx(M, i, j):
    if i > j:
        return
    MSub = M[i:j + 1, i:j + 1]
    return MSub


def getMinor(M, i, j):
    if i > j:
        return
    cofactor = M[range(i) + range(i + 1, M.shape[0]), :]
    cofactor = cofactor[:, range(j) + range(j + 1, M.shape[1])]
    return np.linalg.det(cofactor)


def cofactorDet(M, row):
    detcof = 0
    for k in range(M.shape[1]):
        partial_res = ((-1.) ** (row + k)) * M[row, k] * getMinor(M, row, k)
        detcof += partial_res
        print (row, k), partial_res
    return detcof


def isMtxComplete(M,noedge):
    # return true if mtx M does not contain -1
    if M.min() == noedge: return False
    else: return True


#return true if the principal minor with top-right corner i,j does not contain -1
def isMtxCompleteByIndex(M,i,j,noedge):
    MSub = getSubMtx(M,i,j)
    if MSub.min() == noedge: return False
    else: return True


#return top right index of lexicographic first principal minor with missing top-right value
#we call the principal minor with one missing value in top right corner a singleton
def getSmallestSingleton(M, noedge):
    (n,m) = np.shape(M)
    for j in range(0,m):
        for i in range(0,m-j):
            if M[i,i+j] == noedge:
                return (i,i+j)
    sys.stderr.write('M is complete!\n\n')
    return (0,0)


#return determinant of matrix M
#assumes matrix is square and does not check
def getDet(M):
    return np.linalg.det(M)


#return determinant of submatrix Mij
#assumes matrix is square and does not check
def getDetByIndex(M,i,j):
    return np.linalg.det(getSubMtx(M,i,j))


#return determinant of matrix M
#assumes matrix is square and does not check
def getDetLog(M):
    return np.linalg.slogdet(M)


#return logdeterminant of submatrix Mij
#assumes matrix is square and does not check
def getDetLogByIndex(M,i,j):
    return np.linalg.slogdet(getSubMtx(M,i,j))


#returns value for M_1n = M_n1 so that det(M) is maximized
#assumes only M[1,n] and M[n,1] are missing
def solveMaxDet(M):
    (n,m) = np.shape(M)
    b = M[0,1:n-1]
    c = M[1:n-1,n-1]
    A = M[1:n-1,1:n-1]
    Mn1 = np.dot(np.dot(b,np.linalg.inv(A)),c).item(0)
    #Mn1 = np.dot(np.dot(b,np.linalg.inv(A)),c)
    return Mn1



#returns interval in which  M_1n = M_n1 can be chosen so that det(M) is pos
#assumes only M[1,n] and M[n,1] are missing
#returns tuple (maxDetValue, completionInterval) (float, list)
def solvePosDefInterval(M):
    (n,m) = np.shape(M)
    n = n-1
    if n<2:
        r = np.sqrt(M[0,0]*M[1,1])
        return(0,((-1)*r,r))
    if n == 2:
        B = M[0,0]
        C = M[2,2]
        A = M[1,1]
        b = M[0,1]
        c = M[1,2]
        detB = B
        detC = C
    else:
        n += 1
        B = M[0:n-1,0:n-1]
        C = M[1:n,1:n]
        A = M[1:n-1,1:n-1]
        b = M[0,1:n-1]
        c = M[1:n-1,n-1]
        detB = np.linalg.det(B)
        detC = np.linalg.det(C)
    r = np.sqrt(detB*detC)/np.linalg.det(A)
    r = float(r)
    #center of the interval
    Mn1 = b.dot(np.linalg.inv(A)).dot(c.T)
    I = [Mn1-r,Mn1+r]
    return (Mn1,I)


#returns log of interval in which  M_1n = M_n1 can be chosen so that det(M) is pos
#assumes only M[1,n] and M[n,1] are missing
#returns tuple (maxDetValueLog, completionIntervalLog) (float, list)
def solvePosDefIntervalLog(M):
    (n,m) = np.shape(M)
    if n==2:
        r = np.sqrt(M[0,0]*M[1,1])
        return(0,((-1)*r,r))
    B = M[0:n-1,0:n-1]
    C = M[1:n,1:n]
    A = M[1:n-1,1:n-1]
    logdetB = np.linalg.slogdet(B)[1]
    logdetC = np.linalg.slogdet(C)[1]
    logdetA = np.linalg.slogdet(A)[1]
    logr = 1/2*logdetB+ 1/2*logdetC + (-1)*logdetA
    #center of the interval
    Mn1 = solveMaxDet(M)
    r = np.exp(logr)
    I = [Mn1-r,Mn1+r]
    return (Mn1,I)


def _get_largest_submatrix(partial, i, j):
    if i > j:
        i, j = j, i
    cols = np.where(~np.isnan(partial[i, :]))[0]
    rows = np.where(~np.isnan(partial[j, :]))[0]
    keep = [i]
    for x in set(cols).intersection(set(rows)):
        keep.append(x)

    keep.append(j)
    return partial[keep, :][:, keep]
