def strategy_inner_order(mask):
    """Given the mask of a partial matrix (unknown entries correspond to ones,
    known entries to zeros), return an ordering on the elements such that unkowns
    that are dependent on many other unknowns have high ranking.
    """

    weights = {}
    for j in range(mask.shape[0]):
        for i in range(j):
            if mask[i, j] == 0:
                continue
            weights[(i, j)] = mask[i:j + 1, i:j + 1].sum().sum() / 2. \
                - mask[i, j]
    strategy_ranking = [(i, j)
                        for (i, j), val in
                        sorted(weights.items(), key=lambda x: (x[1], x[0][0],
                                                               x[0][1]))]
    strategy_ranking.reverse()
    return strategy_ranking


def strategy_outer_order(mask):
    """Given the mask of a partial matrix (unknown entries correspond to ones,
    known entries to zeros), return an ordering on the elements such that
    unkowns on which many other unknowns depend have high ranking.
    """

    weights = {}
    for j in range(mask.shape[0]):
        for i in range(j):
            if mask[i, j] == 0:
                continue
            weights[(i, j)] = mask[:i + 1, j:].sum().sum() - mask[i, j]
    strategy_ranking = [(i, j)
                        for (i, j), val in
                        sorted(weights.items(), key=lambda x: (x[1], x[0][0],
                                                               x[0][1]))]
    strategy_ranking.reverse()
    return strategy_ranking


